/*
 * ExtMediaBSDClient class defintion
 *
 * Copyright (C) 2008-2010 Charles Davis
 *
 * This file is dual-licensed under the LGPL and BSD 3-clause license.
 * It is part of the XNU Kernel Extras project, which aims to extend the
 * XNU kernel beyond what is provided by Apple.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 *
 * ============================================================================
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *  3. Neither the name of the author nor the names of any
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "ExtMediaBSDClient.h"
#include "ExtSCSITaskBuilder.h"
#include "diskio_private.h"
#include <libkern/OSMalloc.h>
#include <IOKit/storage/IOBlockStorageDriver.h>
#include <IOKit/scsi/IOSCSIPrimaryCommandsDevice.h>
#include <IOKit/scsi/IOSCSIProtocolServices.h>
#include <IOKit/scsi/SCSITask.h>
#include <IOKit/IOBSD.h>
#include <IOKit/IOLocks.h>
#include <IOKit/IOLib.h>
#include <sys/errno.h>
#include <sys/kauth.h>
#include <sys/vnode.h>
#include <sys/proc.h>

static net_sf_driver_ExtSCSITaskBuilder *builder = NULL;

/* Helper function to prepare a buffer */
static IOMemoryDescriptor *prepare_buffer( user_addr_t address,
                                           uint32_t length,
                                           IODirection dir,
                                           proc_t proc )
{
    IOMemoryDescriptor *buffer = NULL;
    if (address && length)
    {
        buffer = IOMemoryDescriptor::withAddressRange(
            address,
            length,
            dir,
            proc == kernproc ? kernel_task : current_task() );
    }
    if (buffer)
    {
        if (buffer->prepare() != kIOReturnSuccess)
        {
            buffer->release();
            buffer = NULL;
        }
    }
    return buffer;
}

/* Helper function to complete a buffer */
static void complete_buffer( IOMemoryDescriptor *buffer )
{
    buffer->complete();
    buffer->release();
}

/* Helper function to get the whole media object from a partition media */
static IOMedia *get_whole_media( IOMedia *media )
{
    unsigned depth = 1;
    IOService *service;
    for (service = media; service; service = service->getProvider())
    {
        IOMedia *current;
        if (!(current = OSDynamicCast( IOMedia, service ))) continue;
        if (current->isWhole()) return current;
        depth++;
    }
    if (depth == 1)
        IOLog( "%s: Media doesn't have a whole media object!\n", media->getName() );
    return NULL;
}

/* Callback for completed SCSITasks */
static void task_complete( SCSITaskIdentifier task )
{
    // The original thread blocked on the task. So let's unblock it.
    thread_wakeup( task );
}

/* Helper function to check permissions */
static errno_t auth_scsi_cmd( IOMediaBSDClient *client )
{
    static OSMallocTag tag = NULL;
    // Check permissions.
    // First we have to get a vnode to our device file.
    // Technically, we should be looking up the location of the devfs,
    // and using that instead of hardcoding /dev, but assuming that devfs
    // is mounted at /dev is a pretty safe assumption which simplifies
    // the code.
    if (!tag)
        tag = OSMalloc_Tagalloc( "net.sf.driver.ExtMediaBSDClient", 0 );
    if (!tag)
    {
        IOLog( "%s: Couldn't allocate tag.\n", client->getName() );
        return ENOMEM;
    }
    OSString *bsd_name = OSDynamicCast( OSString, client->getProvider()->getProperty( kIOBSDNameKey ) );
    size_t pathlen = sizeof("/dev/r") + bsd_name->getLength();
    char *path = (char *)OSMalloc( pathlen, tag );
    if (!path)
    {
        IOLog( "%s: Out of memory!\n", client->getName() );
        return ENOMEM;
    }
    strlcpy( path, "/dev/r", pathlen );
    strlcpy( path+6, bsd_name->getCStringNoCopy(), pathlen-5 );
    IOLog( "%s: Reconstructed vnode path is %s.\n", client->getName(), path );
    // Now that we have the path, we can get a vnode for it.
    vfs_context_t vfsctx = vfs_context_create( NULL );
    vnode_t vp;
    errno_t err = vnode_lookup( path, 0, &vp, vfsctx );
    OSFree( path, pathlen, tag );
    if (err)
    {
        IOLog( "%s: vnode_lookup() failed, error %d.\n", client->getName(), err );
        return err;
    }
    // Now that we have a vnode, we can use vnode_authorize to check the
    // current process's credentials against it to see if we're allowed
    // to send a SCSI command.
    err = vnode_authorize( vp, NULL, KAUTH_VNODE_READ_DATA|KAUTH_VNODE_WRITE_DATA, vfsctx );
    vnode_put( vp );
    if (err)
    {
        IOLog( "%s: Access denied to user %d.\n", client->getName(), kauth_cred_getuid( vfs_context_ucred( vfsctx ) ) );
        vfs_context_rele( vfsctx );
        return err;
    }
    vfs_context_rele( vfsctx );
    return 0;
}

/* Helper functions for executing a SCSI command */
int exec_scsi_cmd32( IOMediaBSDClient *client, dk_scsi_command_32_t *data, proc_t proc )
{
    // Verify that no reserved bits are set.
    if (data->reserved0144[0] || data->reserved0144[1])
        return EINVAL;
    // Is this command authorized?
    errno_t err = auth_scsi_cmd( client );
    if (err) return err;
    // We need to get the IOSCSIPrimaryCommandsDevice. If it's present, it's
    // above the IOBlockStorageDevice (which is above the
    // IOBlockStorageDriver).
    IOSCSIPrimaryCommandsDevice *scsidev = OSDynamicCast( IOSCSIPrimaryCommandsDevice, get_whole_media( client->getProvider() )->getProvider()->getProvider()->getProvider() );
    if (!scsidev)
    {
        IOLog( "%s: Not a SCSI deivce; can't execute DKIOCSCSICOMMAND.\n", client->getName() );
        return ENOTTY;
    }
    // Now we take the info we got from userland and build a SCSITask
    // command.
    if (!builder)
        builder = OSTypeAlloc( net_sf_driver_ExtSCSITaskBuilder );
    if (!builder)
    {
        IOLog( "%s: Couldn't allocate SCSITaskBuilder.\n", client->getName() );
        return ENOMEM;
    }
    SCSITaskIdentifier task = builder->GetSCSITask();
    if (!task)
    {
        IOLog( "%s: Didn't get a SCSITask object; can't execute DKIOCSCSICOMMAND.\n", client->getName() );
        return EAGAIN;
    }
    IOMemoryDescriptor *buffer = NULL;
    if (data->direction != kSCSIDataTransfer_NoDataTransfer)
    {
        buffer = prepare_buffer(
            data->buffer,
            data->bufferSize,
            data->direction == kSCSIDataTransfer_FromInitiatorToTarget ? kIODirectionOut : kIODirectionIn,
            proc );
        if (!buffer)
        {
            IOLog( "%s: Couldn't get memory descriptor for buffer.\n", client->getName() );
            builder->ReleaseSCSITask( task );
            return EFAULT;
        }
        builder->SetDataBuffer( task, buffer );
        builder->SetRequestedDataTransferCount( task, data->bufferSize );
    }
    builder->SetTaskAttribute( task, (SCSITaskAttribute)data->attribute );
    builder->SetTimeoutDuration( task, data->timeout );
    builder->SetDataTransferDirection( task, data->direction );
    if (data->cdbSize <= 6)
        builder->SetCommandDescriptorBlock( task,
                                            data->cdb[0],
                                            data->cdb[1],
                                            data->cdb[2],
                                            data->cdb[3],
                                            data->cdb[4],
                                            data->cdb[5] );
    else if (data->cdbSize <= 10)
        builder->SetCommandDescriptorBlock( task,
                                            data->cdb[0],
                                            data->cdb[1],
                                            data->cdb[2],
                                            data->cdb[3],
                                            data->cdb[4],
                                            data->cdb[5],
                                            data->cdb[6],
                                            data->cdb[7],
                                            data->cdb[8],
                                            data->cdb[9] );
    else if (data->cdbSize <= 12)
        builder->SetCommandDescriptorBlock( task,
                                            data->cdb[0],
                                            data->cdb[1],
                                            data->cdb[2],
                                            data->cdb[3],
                                            data->cdb[4],
                                            data->cdb[5],
                                            data->cdb[6],
                                            data->cdb[7],
                                            data->cdb[8],
                                            data->cdb[9],
                                            data->cdb[10],
                                            data->cdb[11] );
    else
        builder->SetCommandDescriptorBlock( task,
                                            data->cdb[0],
                                            data->cdb[1],
                                            data->cdb[2],
                                            data->cdb[3],
                                            data->cdb[4],
                                            data->cdb[5],
                                            data->cdb[6],
                                            data->cdb[7],
                                            data->cdb[8],
                                            data->cdb[9],
                                            data->cdb[10],
                                            data->cdb[11],
                                            data->cdb[12],
                                            data->cdb[13],
                                            data->cdb[14],
                                            data->cdb[15] );
    // Set the task completion callback so the stack knows what to do when
    // it's done.
    builder->SetTaskCompletionCallback( task, task_complete );
    // Now that we've built the task, it's time to execute it.
    scsidev->ExecuteCommand( task );
    // Wait for it...
    assert_wait( task, THREAD_UNINT );
    thread_block( THREAD_CONTINUE_NULL );
    // Done.
    data->response = builder->GetServiceResponse( task );
    data->status = builder->GetTaskStatus( task );
    if (buffer)
    {
        data->bufferSize = builder->GetRealizedDataTransferCount( task );
        complete_buffer( buffer );
    }
    IOMemoryDescriptor *sense = prepare_buffer( data->sense,
                                                data->senseLen,
                                                kIODirectionIn,
                                                proc );
    if (sense)
    {
        SCSI_Sense_Data sense_buf;
        if (builder->GetAutoSenseData( task, &sense_buf, data->senseLen ))
            sense->writeBytes( 0, &sense_buf, data->senseLen );
        complete_buffer( sense );
    }
    // Task complete. Callers are expected to handle SCSI errors on their
    // own.
    builder->ReleaseSCSITask( task );
    return 0;
}

/* Helper functions for executing a SCSI command */
int exec_scsi_cmd64( IOMediaBSDClient *client, dk_scsi_command_64_t *data, proc_t proc )
{
    // Verify that no reserved bits are set.
    if (data->reserved0144[0] || data->reserved0144[1])
        return EINVAL;
    // Is this command authorized?
    errno_t err = auth_scsi_cmd( client );
    if (err) return err;
    // We need to get the IOSCSIPrimaryCommandsDevice. If it's present, it's
    // above the IOBlockStorageDevice (which is above the
    // IOBlockStorageDriver).
    IOSCSIPrimaryCommandsDevice *scsidev = OSDynamicCast( IOSCSIPrimaryCommandsDevice, get_whole_media( client->getProvider() )->getProvider()->getProvider()->getProvider() );
    if (!scsidev)
    {
        IOLog( "%s: Not a SCSI deivce; can't execute DKIOCSCSICOMMAND.\n", client->getName() );
        return ENOTTY;
    }
    // Now we take the info we got from userland and build a SCSITask
    // command.
    if (!builder)
        builder = OSTypeAlloc( net_sf_driver_ExtSCSITaskBuilder );
    if (!builder)
    {
        IOLog( "%s: Couldn't allocate SCSITaskBuilder.\n", client->getName() );
        return ENOMEM;
    }
    SCSITaskIdentifier task = builder->GetSCSITask();
    if (!task)
    {
        IOLog( "%s: Didn't get a SCSITask object; can't execute DKIOCSCSICOMMAND.\n", client->getName() );
        return EAGAIN;
    }
    IOMemoryDescriptor *buffer = NULL;
    if (data->direction != kSCSIDataTransfer_NoDataTransfer)
    {
        buffer = prepare_buffer(
            data->buffer,
            data->bufferSize,
            data->direction == kSCSIDataTransfer_FromInitiatorToTarget ? kIODirectionOut : kIODirectionIn,
            proc );
        if (!buffer)
        {
            IOLog( "%s: Couldn't get memory descriptor for buffer.\n", client->getName() );
            builder->ReleaseSCSITask( task );
            return EFAULT;
        }
        builder->SetDataBuffer( task, buffer );
        builder->SetRequestedDataTransferCount( task, data->bufferSize );
    }
    builder->SetTaskAttribute( task, (SCSITaskAttribute)data->attribute );
    builder->SetTimeoutDuration( task, data->timeout );
    builder->SetDataTransferDirection( task, data->direction );
    if (data->cdbSize <= 6)
        builder->SetCommandDescriptorBlock( task,
                                            data->cdb[0],
                                            data->cdb[1],
                                            data->cdb[2],
                                            data->cdb[3],
                                            data->cdb[4],
                                            data->cdb[5] );
    else if (data->cdbSize <= 10)
        builder->SetCommandDescriptorBlock( task,
                                            data->cdb[0],
                                            data->cdb[1],
                                            data->cdb[2],
                                            data->cdb[3],
                                            data->cdb[4],
                                            data->cdb[5],
                                            data->cdb[6],
                                            data->cdb[7],
                                            data->cdb[8],
                                            data->cdb[9] );
    else if (data->cdbSize <= 12)
        builder->SetCommandDescriptorBlock( task,
                                            data->cdb[0],
                                            data->cdb[1],
                                            data->cdb[2],
                                            data->cdb[3],
                                            data->cdb[4],
                                            data->cdb[5],
                                            data->cdb[6],
                                            data->cdb[7],
                                            data->cdb[8],
                                            data->cdb[9],
                                            data->cdb[10],
                                            data->cdb[11] );
    else
        builder->SetCommandDescriptorBlock( task,
                                            data->cdb[0],
                                            data->cdb[1],
                                            data->cdb[2],
                                            data->cdb[3],
                                            data->cdb[4],
                                            data->cdb[5],
                                            data->cdb[6],
                                            data->cdb[7],
                                            data->cdb[8],
                                            data->cdb[9],
                                            data->cdb[10],
                                            data->cdb[11],
                                            data->cdb[12],
                                            data->cdb[13],
                                            data->cdb[14],
                                            data->cdb[15] );
    // Set the task completion callback so the stack knows what to do when
    // it's done.
    builder->SetTaskCompletionCallback( task, task_complete );
    // Now that we've built the task, it's time to execute it.
    scsidev->ExecuteCommand( task );
    // Wait for it...
    assert_wait( task, THREAD_UNINT );
    thread_block( THREAD_CONTINUE_NULL );
    // Done.
    data->response = builder->GetServiceResponse( task );
    data->status = builder->GetTaskStatus( task );
    if (buffer)
    {
        data->bufferSize = builder->GetRealizedDataTransferCount( task );
        complete_buffer( buffer );
    }
    IOMemoryDescriptor *sense = prepare_buffer( data->sense,
                                                data->senseLen,
                                                kIODirectionIn,
                                                proc );
    if (sense)
    {
        SCSI_Sense_Data sense_buf;
        if (builder->GetAutoSenseData( task, &sense_buf, data->senseLen ))
            sense->writeBytes( 0, &sense_buf, data->senseLen );
        complete_buffer( sense );
    }
    // Task complete. Callers are expected to handle SCSI errors on their
    // own.
    builder->ReleaseSCSITask( task );
    return 0;
}

/* ExtMediaBSDClient class implementation */
OSDefineMetaClassAndStructors( net_sf_driver_ExtMediaBSDClient, IOMediaBSDClient );

#define super IOMediaBSDClient
int net_sf_driver_ExtMediaBSDClient::ioctl( dev_t dev, u_long cmd, caddr_t data,
                                            int flags, proc_t task )
{
    switch (cmd)
    {
    case DKIOCPREVENT:
        if (getProvider()->getProperty( kIOMediaRemovableKey ) == kOSBooleanTrue)
        {
            IOBlockStorageDriver *drvobj = OSDynamicCast( IOBlockStorageDriver, getWholeMedia( getProvider() )->getProvider() );
            if (drvobj)
                return getProvider()->errnoFromReturn( drvobj->lockMedia( true ) );
            else
            {
                IOLog( "%s: No IOBlockStorageDriver for DKIOCPREVENT.\n", getName() );
                return EIO;
            }
        }
        else return 0;
    case DKIOCALLOW:
        if (getProvider()->getProperty( kIOMediaRemovableKey ) == kOSBooleanTrue)
        {
            IOBlockStorageDriver *drvobj = OSDynamicCast( IOBlockStorageDriver, getWholeMedia( getProvider() )->getProvider() );
            if (drvobj)
                return getProvider()->errnoFromReturn( drvobj->lockMedia( false ) );
            else
            {
                IOLog( "%s: No IOBlockStorageDriver for DKIOCALLOW.\n", getName() );
                return EIO;
            }
        }
        else return 0;
    case DKIOCSCSICOMMAND32:
        return exec_scsi_cmd32( this, (dk_scsi_command_32_t *)data, task );
    case DKIOCSCSICOMMAND64:
        return exec_scsi_cmd64( this, (dk_scsi_command_64_t *)data, task );
    default:
        return super::ioctl( dev, cmd, data, flags, task );
    }
}

OSMetaClassDefineReservedUnused( net_sf_driver_ExtMediaBSDClient, 0 );
